#!/usr/bin/env python
#
###############################################################################
#   Copyright (C) 2017 Mike Zingman N4IRR
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
###############################################################################

'''
'''
from binascii import b2a_hex as ahex
from bitarray import bitarray
from bitstring import BitArray
from bitstring import BitString

__author__     = 'Mike Zingman, N4IRR and Cortney T. Buffington, N0MJS'
__copyright__  = 'Copyright (c) 2017 Mike Zingman N4IRR'
__credits__    = 'Cortney T. Buffington, N0MJS; Colin Durbridge, G4EML, Steve Zingman, N4IRS; Jonathan Naylor, G4KLX; Hans Barthen, DL5DI; Torsten Shultze, DG1HT'
__license__    = 'GNU GPLv3'
__maintainer__ = 'Cort Buffington, N0MJS'
__email__      = 'n0mjs@me.com'

##
# DMR AMBE interleave schedule
##
rW = [
      0, 1, 0, 1, 0, 1,
      0, 1, 0, 1, 0, 1,
      0, 1, 0, 1, 0, 1,
      0, 1, 0, 1, 0, 2,
      0, 2, 0, 2, 0, 2,
      0, 2, 0, 2, 0, 2
      ]

rX = [
      23, 10, 22, 9, 21, 8,
      20, 7, 19, 6, 18, 5,
      17, 4, 16, 3, 15, 2,
      14, 1, 13, 0, 12, 10,
      11, 9, 10, 8, 9, 7,
      8, 6, 7, 5, 6, 4
      ]

rY = [
      0, 2, 0, 2, 0, 2,
      0, 2, 0, 3, 0, 3,
      1, 3, 1, 3, 1, 3,
      1, 3, 1, 3, 1, 3,
      1, 3, 1, 3, 1, 3,
      1, 3, 1, 3, 1, 3
      ]

rZ = [
      5, 3, 4, 2, 3, 1,
      2, 0, 1, 13, 0, 12,
      22, 11, 21, 10, 20, 9,
      19, 8, 18, 7, 17, 6,
      16, 5, 15, 4, 14, 3,
      13, 2, 12, 1, 11, 0
      ]


# This function calculates [23,12] Golay codewords.
# The format of the returned longint is [checkbits(11),data(12)].
def golay2312(cw):
    POLY = 0xAE3                #/* or use the other polynomial, 0xC75 */
    cw = cw & 0xfff             # Strip off check bits and only use data
    c = cw                      #/* save original codeword */
    for i in range(1,13):       #/* examine each data bit */
        if (cw & 1):            #/* test data bit */
            cw = cw ^ POLY      #/* XOR polynomial */
        cw = cw >> 1            #/* shift intermediate result */
    return((cw << 12) | c)      #/* assemble codeword */

# This function checks the overall parity of codeword cw.
# If parity is even, 0 is returned, else 1.
def parity(cw):
    #/* XOR the bytes of the codeword */
    p = cw & 0xff
    p = p ^ ((cw >> 8) & 0xff)
    p = p ^ ((cw >> 16) & 0xff)
    
    #/* XOR the halves of the intermediate result */
    p = p ^ (p >> 4)
    p = p ^ (p >> 2)
    p = p ^ (p >> 1)
    
    #/* return the parity result */
    return(p & 1)

# Demodulate ambe frame (C1)
# Frame is an array [4][24]
def demodulateAmbe3600x2450(ambe_fr):
    pr = [0] * 115
    foo = 0

    # create pseudo-random modulator
    for i in range(23, 11, -1):
        foo = foo << 1
        foo = foo | ambe_fr[0][i]
    pr[0] = (16 * foo)
    for i in range(1, 24):
        pr[i] = (173 * pr[i - 1]) + 13849 - (65536 * (((173 * pr[i - 1]) + 13849) // 65536))
    for i in range(1, 24):
        pr[i] = pr[i] // 32768

    # demodulate ambe_fr with pr
    k = 1
    for j in range(22, -1, -1):
        ambe_fr[1][j] = ((ambe_fr[1][j]) ^ pr[k])
        k = k + 1
    return ambe_fr  # Pass it back since there is no pass by reference

def eccAmbe3600x2450Data(ambe_fr):
    ambe = bitarray()
    
    # just copy C0
    for j in range(23, 11, -1):
        ambe.append(ambe_fr[0][j])
    
#        # ecc and copy C1
#        gin = 0
#        for j in range(23):
#            gin = (gin << 1) | ambe_fr[1][j]
#
#        gout = BitArray(hex(golay2312(gin)))
#        for j in range(22, 10, -1):
#            ambe[bitIndex] = gout[j]
#            bitIndex += 1
    for j in range(22, 10, -1):
        ambe.append(ambe_fr[1][j])

    # just copy C2
    for j in range(10, -1, -1):
        ambe.append(ambe_fr[2][j])

    # just copy C3
    for j in range(13, -1, -1):
        ambe.append(ambe_fr[3][j])

    return ambe

# Convert a 49 bit raw AMBE frame into a deinterleaved structure (ready for decode by AMBE3000)
def convert49BitAmbeTo72BitFrames( ambe_d ):
    index = 0
    ambe_fr = [[None for x in range(24)] for y in range(4)]

    #Place bits into the 4x24 frames.  [bit0...bit23]
    #fr0: [P e10 e9 e8 e7 e6 e5 e4 e3 e2 e1 e0 11 10 9 8 7 6 5 4 3 2 1 0]
    #fr1: [e10 e9 e8 e7 e6 e5 e4 e3 e2 e1 e0 23 22 21 20 19 18 17 16 15 14 13 12 xx]
    #fr2: [34 33 32 31 30 29 28 27 26 25 24 x x x x x x x x x x x x x]
    #fr3: [48 47 46 45 44 43 42 41 40 39 38 37 36 35 x x x x x x x x x x]

    # ecc and copy C0: 12bits + 11ecc + 1 parity
    # First get the 12 bits that actually exist
    # Then calculate the golay codeword
    # And then add the parity bit to get the final 24 bit pattern

    tmp = 0
    for i in range(11, -1, -1):      #grab the 12 MSB
        tmp = (tmp << 1) | ambe_d[i]
    tmp = golay2312(tmp)               #Generate the 23 bit result
    parityBit = parity(tmp)
    tmp = tmp | (parityBit << 23)               #And create a full 24 bit value
    for i in range(23, -1, -1):
        ambe_fr[0][i] = (tmp & 1)
        tmp = tmp >> 1

    # C1: 12 bits + 11ecc (no parity)
    tmp = 0
    for i in range(23,11, -1) :         #grab the next 12 bits
        tmp = (tmp << 1) | ambe_d[i]
    tmp = golay2312(tmp)                    #Generate the 23 bit result
    for j in range(22, -1, -1):
        ambe_fr[1][j] = (tmp & 1)
        tmp = tmp >> 1;

    #C2: 11 bits (no ecc)
    for j in range(10, -1, -1):
        ambe_fr[2][j] = ambe_d[34 - j]

    #C3: 14 bits (no ecc)
    for j in range(13, -1, -1):
        ambe_fr[3][j] = ambe_d[48 - j];

    return ambe_fr
        
def interleave(ambe_fr):
    bitIndex = 0
    w = 0
    x = 0
    y = 0
    z = 0
    data = bytearray(9)
    for i in range(36):
        bit1 = ambe_fr[rW[w]][rX[x]] # bit 1
        bit0 = ambe_fr[rY[y]][rZ[z]] # bit 0


        data[bitIndex // 8] = ((data[bitIndex // 8] << 1) & 0xfe) | (1 if (bit1 == 1) else 0)
        bitIndex += 1

        data[bitIndex // 8] = ((data[bitIndex // 8] << 1) & 0xfe) | (1 if (bit0 == 1) else 0)
        bitIndex += 1

        w += 1
        x += 1
        y += 1 
        z += 1
    return data

def deinterleave(data):
    
    ambe_fr = [[None for x in range(24)] for y in range(4)]

    bitIndex = 0
    w = 0
    x = 0
    y = 0
    z = 0
    for i in range(36):
        bit1 = 1 if data[bitIndex] else 0
        bitIndex += 1

        bit0 = 1 if data[bitIndex] else 0
        bitIndex += 1

        ambe_fr[rW[w]][rX[x]] = bit1; # bit 1
        ambe_fr[rY[y]][rZ[z]] = bit0; # bit 0

        w += 1
        x += 1
        y += 1
        z += 1

    return ambe_fr

def convert72BitTo49BitAMBE( ambe72 ):
    ambe_fr = deinterleave(ambe72)                     # take 72 bit ambe and lay it out in C0-C3
    ambe_fr = demodulateAmbe3600x2450(ambe_fr)         # demodulate C1
    ambe49 = eccAmbe3600x2450Data(ambe_fr)             # pick out the 49 bits of raw ambe
    return ambe49

def convert49BitTo72BitAMBE( ambe49 ):
    ambe_fr = convert49BitAmbeTo72BitFrames(ambe49)    # take raw ambe 49 + ecc and place it into C0-C3
    ambe_fr = demodulateAmbe3600x2450(ambe_fr)         # demodulate C1
    ambe72 = interleave(ambe_fr);                      # Re-interleave it, returning 72 bits
    return ambe72

def testit():
    ambe72 = BitArray('0xACAA40200044408080')    #silence frame
    print('ambe72=',ambe72)
    
    ambe49 = convert72BitTo49BitAMBE(ambe72)
    print('ambe49=',ahex(ambe49.tobytes()))

    ambe72 = convert49BitTo72BitAMBE(ambe49)
    print('ambe72=',ahex(ambe72))
    
#------------------------------------------------------------------------------
# Used to execute the module directly to run built-in tests
#------------------------------------------------------------------------------

if __name__ == '__main__':
    testit()



#!/usr/bin/env python
#
###############################################################################
#   Copyright (C) 2016-2018  Cortney T. Buffington, N0MJS <n0mjs@me.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
###############################################################################


# Does anybody read this stuff? There's a PEP somewhere that says I should do this.
__author__     = 'Cortney T. Buffington, N0MJS'
__copyright__  = 'Copyright (c) 2016-2018 Cortney T. Buffington, N0MJS and the K0USY Group'
__credits__    = 'DSD'
__license__    = 'GNU GPLv3'
__maintainer__ = 'Cort Buffington, N0MJS'
__email__      = 'n0mjs@me.com'

inter_X = (                                                  
  23,  5, 10,  3, 22,  4, 9,  2, 21,  3, 8, 1, 20,  2, 7, 0, 19,  1, 6, 13, 18,  0,  5, 12,
  17, 22,  4, 11, 16, 21, 3, 10, 15, 20, 2, 9, 14, 19, 1, 8, 13, 18, 0,  7, 12, 17, 10, 6,
  11, 16,  9,  5, 10, 15, 8,  4,  9, 14, 7, 3,  8, 13, 6, 2,  7, 12, 5,  1,  6, 11,  4, 0
)

inter_W = (                                                  
  0, 0, 1, 2, 0, 0, 1, 2, 0, 0, 1, 2, 0, 0, 1, 2, 0, 0, 1, 3, 0, 0, 1, 3,       
  0, 1, 1, 3, 0, 1, 1, 3, 0, 1, 1, 3, 0, 1, 1, 3, 0, 1, 1, 3, 0, 1, 2, 3,       
  0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3        
)

if __name__ == '__main__':
    print(inter_X[13])


#!/usr/bin/env python
#
###############################################################################
#   Copyright (C) 2016  Cortney T. Buffington, N0MJS <n0mjs@me.com>
#   Copyright (C) 2015 by Jonathan Naylor G4KLX
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
###############################################################################



START_MASK = [0x96, 0x96, 0x96]
END_MASK = [0x99, 0x99, 0x99]
NUM_BYTES = 9
NPAR = 3;
POLY= [64, 56, 14, 1, 0, 0, 0, 0, 0, 0, 0, 0]

EXP_TABLE = (
	0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1D, 0x3A, 0x74, 0xE8, 0xCD, 0x87, 0x13, 0x26,
	0x4C, 0x98, 0x2D, 0x5A, 0xB4, 0x75, 0xEA, 0xC9, 0x8F, 0x03, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0,
	0x9D, 0x27, 0x4E, 0x9C, 0x25, 0x4A, 0x94, 0x35, 0x6A, 0xD4, 0xB5, 0x77, 0xEE, 0xC1, 0x9F, 0x23,
	0x46, 0x8C, 0x05, 0x0A, 0x14, 0x28, 0x50, 0xA0, 0x5D, 0xBA, 0x69, 0xD2, 0xB9, 0x6F, 0xDE, 0xA1,
	0x5F, 0xBE, 0x61, 0xC2, 0x99, 0x2F, 0x5E, 0xBC, 0x65, 0xCA, 0x89, 0x0F, 0x1E, 0x3C, 0x78, 0xF0,
	0xFD, 0xE7, 0xD3, 0xBB, 0x6B, 0xD6, 0xB1, 0x7F, 0xFE, 0xE1, 0xDF, 0xA3, 0x5B, 0xB6, 0x71, 0xE2,
	0xD9, 0xAF, 0x43, 0x86, 0x11, 0x22, 0x44, 0x88, 0x0D, 0x1A, 0x34, 0x68, 0xD0, 0xBD, 0x67, 0xCE,
	0x81, 0x1F, 0x3E, 0x7C, 0xF8, 0xED, 0xC7, 0x93, 0x3B, 0x76, 0xEC, 0xC5, 0x97, 0x33, 0x66, 0xCC,
	0x85, 0x17, 0x2E, 0x5C, 0xB8, 0x6D, 0xDA, 0xA9, 0x4F, 0x9E, 0x21, 0x42, 0x84, 0x15, 0x2A, 0x54,
	0xA8, 0x4D, 0x9A, 0x29, 0x52, 0xA4, 0x55, 0xAA, 0x49, 0x92, 0x39, 0x72, 0xE4, 0xD5, 0xB7, 0x73,
	0xE6, 0xD1, 0xBF, 0x63, 0xC6, 0x91, 0x3F, 0x7E, 0xFC, 0xE5, 0xD7, 0xB3, 0x7B, 0xF6, 0xF1, 0xFF,
	0xE3, 0xDB, 0xAB, 0x4B, 0x96, 0x31, 0x62, 0xC4, 0x95, 0x37, 0x6E, 0xDC, 0xA5, 0x57, 0xAE, 0x41,
	0x82, 0x19, 0x32, 0x64, 0xC8, 0x8D, 0x07, 0x0E, 0x1C, 0x38, 0x70, 0xE0, 0xDD, 0xA7, 0x53, 0xA6,
	0x51, 0xA2, 0x59, 0xB2, 0x79, 0xF2, 0xF9, 0xEF, 0xC3, 0x9B, 0x2B, 0x56, 0xAC, 0x45, 0x8A, 0x09,
	0x12, 0x24, 0x48, 0x90, 0x3D, 0x7A, 0xF4, 0xF5, 0xF7, 0xF3, 0xFB, 0xEB, 0xCB, 0x8B, 0x0B, 0x16,
	0x2C, 0x58, 0xB0, 0x7D, 0xFA, 0xE9, 0xCF, 0x83, 0x1B, 0x36, 0x6C, 0xD8, 0xAD, 0x47, 0x8E, 0x01,
	0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1D, 0x3A, 0x74, 0xE8, 0xCD, 0x87, 0x13, 0x26, 0x4C,
	0x98, 0x2D, 0x5A, 0xB4, 0x75, 0xEA, 0xC9, 0x8F, 0x03, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0x9D,
	0x27, 0x4E, 0x9C, 0x25, 0x4A, 0x94, 0x35, 0x6A, 0xD4, 0xB5, 0x77, 0xEE, 0xC1, 0x9F, 0x23, 0x46,
	0x8C, 0x05, 0x0A, 0x14, 0x28, 0x50, 0xA0, 0x5D, 0xBA, 0x69, 0xD2, 0xB9, 0x6F, 0xDE, 0xA1, 0x5F,
	0xBE, 0x61, 0xC2, 0x99, 0x2F, 0x5E, 0xBC, 0x65, 0xCA, 0x89, 0x0F, 0x1E, 0x3C, 0x78, 0xF0, 0xFD,
	0xE7, 0xD3, 0xBB, 0x6B, 0xD6, 0xB1, 0x7F, 0xFE, 0xE1, 0xDF, 0xA3, 0x5B, 0xB6, 0x71, 0xE2, 0xD9,
	0xAF, 0x43, 0x86, 0x11, 0x22, 0x44, 0x88, 0x0D, 0x1A, 0x34, 0x68, 0xD0, 0xBD, 0x67, 0xCE, 0x81,
	0x1F, 0x3E, 0x7C, 0xF8, 0xED, 0xC7, 0x93, 0x3B, 0x76, 0xEC, 0xC5, 0x97, 0x33, 0x66, 0xCC, 0x85,
	0x17, 0x2E, 0x5C, 0xB8, 0x6D, 0xDA, 0xA9, 0x4F, 0x9E, 0x21, 0x42, 0x84, 0x15, 0x2A, 0x54, 0xA8,
	0x4D, 0x9A, 0x29, 0x52, 0xA4, 0x55, 0xAA, 0x49, 0x92, 0x39, 0x72, 0xE4, 0xD5, 0xB7, 0x73, 0xE6,
	0xD1, 0xBF, 0x63, 0xC6, 0x91, 0x3F, 0x7E, 0xFC, 0xE5, 0xD7, 0xB3, 0x7B, 0xF6, 0xF1, 0xFF, 0xE3,
	0xDB, 0xAB, 0x4B, 0x96, 0x31, 0x62, 0xC4, 0x95, 0x37, 0x6E, 0xDC, 0xA5, 0x57, 0xAE, 0x41, 0x82,
	0x19, 0x32, 0x64, 0xC8, 0x8D, 0x07, 0x0E, 0x1C, 0x38, 0x70, 0xE0, 0xDD, 0xA7, 0x53, 0xA6, 0x51,
	0xA2, 0x59, 0xB2, 0x79, 0xF2, 0xF9, 0xEF, 0xC3, 0x9B, 0x2B, 0x56, 0xAC, 0x45, 0x8A, 0x09, 0x12,
	0x24, 0x48, 0x90, 0x3D, 0x7A, 0xF4, 0xF5, 0xF7, 0xF3, 0xFB, 0xEB, 0xCB, 0x8B, 0x0B, 0x16, 0x2C,
	0x58, 0xB0, 0x7D, 0xFA, 0xE9, 0xCF, 0x83, 0x1B, 0x36, 0x6C, 0xD8, 0xAD, 0x47, 0x8E, 0x01, 0x00
    )

LOG_TABLE = (
	0x00, 0x00, 0x01, 0x19, 0x02, 0x32, 0x1A, 0xC6, 0x03, 0xDF, 0x33, 0xEE, 0x1B, 0x68, 0xC7, 0x4B,
	0x04, 0x64, 0xE0, 0x0E, 0x34, 0x8D, 0xEF, 0x81, 0x1C, 0xC1, 0x69, 0xF8, 0xC8, 0x08, 0x4C, 0x71,
	0x05, 0x8A, 0x65, 0x2F, 0xE1, 0x24, 0x0F, 0x21, 0x35, 0x93, 0x8E, 0xDA, 0xF0, 0x12, 0x82, 0x45,
	0x1D, 0xB5, 0xC2, 0x7D, 0x6A, 0x27, 0xF9, 0xB9, 0xC9, 0x9A, 0x09, 0x78, 0x4D, 0xE4, 0x72, 0xA6,
	0x06, 0xBF, 0x8B, 0x62, 0x66, 0xDD, 0x30, 0xFD, 0xE2, 0x98, 0x25, 0xB3, 0x10, 0x91, 0x22, 0x88,
	0x36, 0xD0, 0x94, 0xCE, 0x8F, 0x96, 0xDB, 0xBD, 0xF1, 0xD2, 0x13, 0x5C, 0x83, 0x38, 0x46, 0x40,
	0x1E, 0x42, 0xB6, 0xA3, 0xC3, 0x48, 0x7E, 0x6E, 0x6B, 0x3A, 0x28, 0x54, 0xFA, 0x85, 0xBA, 0x3D,
	0xCA, 0x5E, 0x9B, 0x9F, 0x0A, 0x15, 0x79, 0x2B, 0x4E, 0xD4, 0xE5, 0xAC, 0x73, 0xF3, 0xA7, 0x57,
	0x07, 0x70, 0xC0, 0xF7, 0x8C, 0x80, 0x63, 0x0D, 0x67, 0x4A, 0xDE, 0xED, 0x31, 0xC5, 0xFE, 0x18,
	0xE3, 0xA5, 0x99, 0x77, 0x26, 0xB8, 0xB4, 0x7C, 0x11, 0x44, 0x92, 0xD9, 0x23, 0x20, 0x89, 0x2E,
	0x37, 0x3F, 0xD1, 0x5B, 0x95, 0xBC, 0xCF, 0xCD, 0x90, 0x87, 0x97, 0xB2, 0xDC, 0xFC, 0xBE, 0x61,
	0xF2, 0x56, 0xD3, 0xAB, 0x14, 0x2A, 0x5D, 0x9E, 0x84, 0x3C, 0x39, 0x53, 0x47, 0x6D, 0x41, 0xA2,
	0x1F, 0x2D, 0x43, 0xD8, 0xB7, 0x7B, 0xA4, 0x76, 0xC4, 0x17, 0x49, 0xEC, 0x7F, 0x0C, 0x6F, 0xF6,
	0x6C, 0xA1, 0x3B, 0x52, 0x29, 0x9D, 0x55, 0xAA, 0xFB, 0x60, 0x86, 0xB1, 0xBB, 0xCC, 0x3E, 0x5A,
	0xCB, 0x59, 0x5F, 0xB0, 0x9C, 0xA9, 0xA0, 0x51, 0x0B, 0xF5, 0x16, 0xEB, 0x7A, 0x75, 0x2C, 0xD7,
	0x4F, 0xAE, 0xD5, 0xE9, 0xE6, 0xE7, 0xAD, 0xE8, 0x74, 0xD6, 0xF4, 0xEA, 0xA8, 0x50, 0x58, 0xAF
    )
      
# multiplication using logarithms
def log_mult(a, b):
    if a == 0 or b == 0:
	    return 0
    x = LOG_TABLE[a]
    y = LOG_TABLE[b]
    z = EXP_TABLE[x + y]
    return z

# Reed-Solomon (12,9) encoder
def encode(_msg):
    assert len(_msg) == 9, 'RS129_encode error: Message not 9 bytes: %s' % print_hex(_msg)
    
    parity = [0x00, 0x00, 0x00]

    for i in range(NUM_BYTES):
        dbyte = _msg[i] ^ parity[NPAR - 1]        
        for j in range(NPAR - 1, 0, -1):
            parity[j] = parity[j - 1] ^ log_mult(POLY[j], dbyte)
        parity[0] = log_mult(POLY[0], dbyte)
    return [parity[2], parity[1], parity[0]]

# Apply DMR XOR LC Header MASK
def lc_header_mask(_parity):
    xor = [0,0,0]
    for i in range(len(_parity)):
        xor[i] = _parity[i] ^ START_MASK[i]
    return xor
    
# Apply DMR XOR LC Terminator MASK
def lc_terminator_mask(_parity):
    xor = [0,0,0]
    for i in range(len(_parity)):
        xor[i] = _parity[i] ^ END_MASK[i]
    return xor

# All Inclusive function to take an LC string and provide the RS129 string to append
def lc_header_encode(_message):
    bin_message = bytearray(_message)
    parity = encode(bin_message)
    masked_parity = lc_header_mask(parity)
    return bytes([masked_parity[0]]) + bytes([masked_parity[1]]) + bytes([masked_parity[2]])
    
# All Inclusive function to take an LC string and provide the RS129 string to append
def lc_terminator_encode(_message):
    bin_message = bytearray(_message)
    parity = encode(bin_message)
    masked_parity = lc_terminator_mask(parity)
    return bytes([masked_parity[0]]) + bytes([masked_parity[1]]) + bytes([masked_parity[2]])
    
    


if __name__ == '__main__':    
    from binascii import b2a_hex as ahex

    # For testing the code
    def print_hex(_list):
        print('[{}]'.format(', '.join(hex(x) for x in _list)))
    
    # Validation Example
    message = b'\x00\x10\x20\x00\x0c\x30\x2f\x9b\xe5'
    parity_should_be = b'\xda\x4d\x5a'
    print('Original Message:            {}'.format(ahex(message)))
    print('Masked Parity Should be:     {}'.format(ahex(parity_should_be)))
    
    parity = lc_header_encode(message)
    print('Calculated Masked Parity is: {}'.format(ahex(parity)))



#!/usr/bin/env python
#
###############################################################################
#   Copyright (C) 2016-2018  Cortney T. Buffington, N0MJS <n0mjs@me.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
###############################################################################


from bitarray import bitarray


__author__     = 'Cortney T. Buffington, N0MJS'
__copyright__  = 'Copyright (c) 2016-2018 Cortney T. Buffington, N0MJS and the K0USY Group'
__credits__    = ''
__license__    = 'GNU GPLv3'
__maintainer__ = 'Cort Buffington, N0MJS'
__email__      = 'n0mjs@me.com'

# Slot Type Data types
DMR_SLT_VHEAD = b'\x01'
DMR_SLT_VTERM = b'\x02'

# Sync patterns used for LC and Voice Burst A packets
BS_VOICE_SYNC = bitarray()
BS_DATA_SYNC  = bitarray()
BS_VOICE_SYNC.frombytes(b'\x75\x5F\xD7\xDF\x75\xF7')
BS_DATA_SYNC.frombytes(b'\xDF\xF5\x7D\x75\xDF\x5D')

SYNC = {
    'BS_VOICE': BS_VOICE_SYNC,
    'BS_DATA':  BS_DATA_SYNC
}

# LC Options - G for Group, U for unit, LC_OPT is legacy for group.
LC_OPT_G = b'\x00\x00\x00'
LC_OPT_U = b'\x03\x00\x00'
LC_OPT = LC_OPT_G

# Precomputed EMB values, where CC always = 1, and PI always = 0
EMB = {
    'BURST_B': bitarray('0001001110010001'),
    'BURST_C': bitarray('0001011101110100'),
    'BURST_D': bitarray('0001011101110100'),
    'BURST_E': bitarray('0001010100000111'),
    'BURST_F': bitarray('0001000111100010')
}

# Precomputed Slot Type values where CC always = 1
SLOT_TYPE = {
    'PI_HEAD':       bitarray('00010000001101100111'),
    'VOICE_LC_HEAD': bitarray('00010001101110001100'),
    'VOICE_LC_TERM': bitarray('00010010101001011001'),
    'CSBK':          bitarray('00010011001010110010'),
    'MBC_HEAD':      bitarray('00010100100111110000'),
    'MBC_CONT':      bitarray('00010101000100011011'),
    'DATA_HEAD':     bitarray('00010110000011001110'),
    '1/2_DATA':      bitarray('00010111100000100101'),
    '3/4_DATA':      bitarray('00011000111010100001'),
    'IDLE':          bitarray('00011001011001001010'),
    '1/1_DATA':      bitarray('00011010011110011111'),
    'RES_1':         bitarray('00011011111101110100'),
    'RES_2':         bitarray('00011100010000110110'),
    'RES_3':         bitarray('00011101110011011101'),
    'RES_4':         bitarray('00011110110100001000'),
    'RES_5':         bitarray('00011111010111100011')
}

# LC infor for first 3 Bytes:
# Byte 1: PF (1),Res(1),FLCO(6) -- Byte 2: FID(8) -- Byte 3: Service Options(8)
LC_VOICE = {
    'FLCO-GRP': bitarray('00000000'),
    'FLCO-USR': bitarray('00000011'),
    'FID-GENC': bitarray('00000000'),
    'FID-MOTO': bitarray('00010000'),
    'SVC-OVCM': bitarray('00100000'),
    'SVC-NONE': bitarray('00000000')
}

# Data Type as defined for use in the Slot Type Message
DATA_TYPE = {
    'PI_HEAD':       bitarray('0000'),
    'VOICE_LC_HEAD': bitarray('0001'),
    'VOICE_LC_TERM': bitarray('0010'),
    'CSBK':          bitarray('0011'),
    'MBC_HEAD':      bitarray('0100'),
    'MBC_CONT':      bitarray('0101'),
    'DATA_HEAD':     bitarray('0110'),
    '1/2_RATE':      bitarray('0111'),
    '3/4_RATE':      bitarray('1000'),
    'IDLE':          bitarray('1001'),
    '1/1_RATE':      bitarray('1010'),
    'RES_1':         bitarray('1011'),
    'RES_2':         bitarray('1100'),
    'RES_3':         bitarray('1101'),
    'RES_4':         bitarray('1110'),
    'RES_5':         bitarray('1111')
}




'''
EMB: CC(4b), PI(1b), LCSS(2b), EMB Parity(9b - QR 16,7,5)
Slot Type: CC(4b), DataType(4), Slot Type Parity(12b - )

'''

#------------------------------------------------------------------------------
# Used to execute the module directly to run built-in tests
#------------------------------------------------------------------------------

if __name__ == '__main__':
    
    from pprint import pprint
    
    pprint(SYNC)
    pprint(EMB)
    pprint(SLOT_TYPE)
    print(LC_OPT)
